"""Module for Kubectl-related functionality."""

from pathlib import Path
from time import sleep

import yaml
from azure.core.exceptions import ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.mgmt.containerservice import ContainerServiceClient

kubeconfig_path = Path("~/.kube/config").expanduser()


def add_cluster_to_kubeconfig(
    credential: DefaultAzureCredential,
    subscription_id: str,
    resource_group_name: str,
    cluster_name: str,
) -> None:
    """Add AKS cluster credentials to kubeconfig.

    Args:
        credential (DefaultAzureCredential): Azure credentials for authentication.
        subscription_id (str): Azure subscription ID.
        resource_group_name (str): Name of the resource group containing the AKS cluster.
        cluster_name (str): Name of the AKS cluster.
        rich_progress (Progress | None, optional): Optional Rich Progress instance for progress tracking. Defaults to None.
        rich_task (TaskID | None, optional): Optional task ID for progress tracking. Defaults to None.

    Returns:
        None

    """  # noqa: E501
    aks_client = ContainerServiceClient(credential, subscription_id)
    for _ in range(20):
        try:
            credentials = aks_client.managed_clusters.list_cluster_admin_credentials(resource_group_name, cluster_name)
            break
        except ResourceNotFoundError:
            sleep(20)
    kubeconfig = credentials.kubeconfigs[0].value.decode("utf-8")

    if kubeconfig_path.exists():
        existing_config = yaml.safe_load(kubeconfig_path.read_text())
        new_config = yaml.safe_load(kubeconfig)
        existing_config["clusters"].extend(new_config["clusters"])
        existing_config["contexts"].extend(new_config["contexts"])
        existing_config["users"].extend(new_config["users"])
        kubeconfig_path.write_text(yaml.safe_dump(existing_config))
    else:
        kubeconfig_path.parent.mkdir(parents=True, exist_ok=True)
        kubeconfig_path.write_text(kubeconfig)


def set_current_context(cluster_name: str) -> None:
    """Set the current context in kubeconfig to the specified AKS cluster.

    Args:
        cluster_name (str): Name of the AKS cluster to set as current context.

    Returns:
        None

    """
    if kubeconfig_path.exists():
        kubeconfig = yaml.safe_load(kubeconfig_path.read_text())
        for context in kubeconfig["contexts"]:
            if cluster_name in context["name"]:
                kubeconfig["current-context"] = context["name"]
                break
        kubeconfig_path.write_text(yaml.safe_dump(kubeconfig))


def remove_from_kubeconfig(resource_group_name: str) -> None:
    """Remove AKS cluster credentials from kubeconfig.

    Args:
        resource_group_name (str): Name of the resource group containing the AKS cluster.

    Returns:
        None

    """
    if kubeconfig_path.exists():
        kubeconfig = yaml.safe_load(kubeconfig_path.read_text())
        kubeconfig["clusters"] = [c for c in kubeconfig["clusters"] if resource_group_name not in c["name"]]
        kubeconfig["contexts"] = [c for c in kubeconfig["contexts"] if resource_group_name not in c["name"]]
        kubeconfig["users"] = [c for c in kubeconfig["users"] if resource_group_name not in c["name"]]
        kubeconfig_path.write_text(yaml.safe_dump(kubeconfig))


def cleanup_kubeconfig() -> None:
    """Clean up kubeconfig by removing all clusters, contexts, and users that start with "kina-".

    Returns:
        None

    """
    if kubeconfig_path.exists():
        kubeconfig = yaml.safe_load(kubeconfig_path.read_text())
        kubeconfig["clusters"] = [c for c in kubeconfig["clusters"] if not c["name"].startswith("kina-")]
        kubeconfig["contexts"] = [c for c in kubeconfig["contexts"] if not c["name"].startswith("kina-")]
        kubeconfig["users"] = [c for c in kubeconfig["users"] if not c["name"].startswith("clusterUser_kina-")]
        kubeconfig_path.write_text(yaml.safe_dump(kubeconfig))



"""Functions to manage Azure Virtual Networks."""

from ipaddress import IPv4Network, ip_network

from azure.identity import DefaultAzureCredential
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.network.models._models_py3 import VirtualNetwork
from rich.progress import Progress, TaskID

_base_cidr = ip_network("10.0.0.0/16")


def generate_cidrs() -> tuple[IPv4Network, IPv4Network, IPv4Network]:
    """Generate CIDR blocks for virtual network, pod network, and service network.

    Returns:
        tuple[IPv4Network, IPv4Network, IPv4Network]: A tuple containing the virtual network, pod network, and service network CIDR blocks.

    """  # noqa: E501
    global _base_cidr  # noqa: PLW0603
    vnet_network = _base_cidr
    pod_network = ip_network((vnet_network.network_address + vnet_network.num_addresses, vnet_network.prefixlen))
    service_network = ip_network((pod_network.network_address + pod_network.num_addresses, pod_network.prefixlen))
    _base_cidr = ip_network(
        (service_network.network_address + service_network.num_addresses, service_network.prefixlen),
    )
    return vnet_network, pod_network, service_network


def create_vnet(
    client: NetworkManagementClient,
    resource_group_name: str,
    region: str,
    vnet_cidr: IPv4Network,
    pod_cidr: IPv4Network,
    svc_cidr: IPv4Network,
) -> VirtualNetwork:
    """Create a virtual network in the specified region.

    Args:
        client (NetworkManagementClient): Azure Network Management client.
        resource_group_name (str): Name of the resource group to create the virtual network in.
        region (str): Azure region to create the virtual network in.
        vnet_cidr (IPv4Network): CIDR block for the virtual network.
        pod_cidr (IPv4Network): CIDR block for the pod network.
        svc_cidr (IPv4Network): CIDR block for the service network.

    Returns:
        VirtualNetwork: The created virtual network object.

    """
    return client.virtual_networks.begin_create_or_update(
        resource_group_name=resource_group_name,
        virtual_network_name=f"{resource_group_name}-{region}",
        parameters={
            "location": region,
            "tags": {
                "pod-network": str(pod_cidr),
                "service-network": str(svc_cidr),
                "dns-service-ip": str(svc_cidr.network_address + 10),
            },
            "address_space": {"address_prefixes": [vnet_cidr]},
        },
    ).result()


def peer_vnets(
    client: NetworkManagementClient,
    subscription_id: str,
    resource_group_name: str,
    vnets: list[str],
) -> None:
    """Peer multiple virtual networks together.

    Args:
        client (NetworkManagementClient): Azure Network Management client.
        subscription_id (str): Azure subscription ID.
        resource_group_name (str): Name of the resource group containing the virtual networks.
        vnets (list[str]): List of virtual network names to be peered.

    """
    for i in range(len(vnets)):
        peers = vnets[:i] + vnets[i + 1 :]
        for peer in peers:
            client.virtual_network_peerings.begin_create_or_update(
                resource_group_name=resource_group_name,
                virtual_network_name=vnets[i],
                virtual_network_peering_name=f"{vnets[i].split('-')[-1]}-to-{peer.split('-')[-1]}",
                virtual_network_peering_parameters={
                    "allow_virtual_network_access": True,
                    "allow_forwarded_traffic": True,
                    "allow_gateway_transit": False,
                    "use_remote_gateways": False,
                    "remote_virtual_network": {
                        "id": f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{peer}",  # noqa: E501
                    },
                },
            ).result()


def create_subnet(
    client: NetworkManagementClient,
    resource_group_name: str,
    vnet_name: str,
    subnet_name: str,
    cidr: str,
) -> None:
    """Create a subnet in the specified virtual network.

    Args:
        client (NetworkManagementClient): Azure Network Management client.
        resource_group_name (str): Name of the resource group containing the virtual network.
        vnet_name (str): Name of the virtual network to create the subnet in.
        subnet_name (str): Name of the subnet to create.
        cidr (str): CIDR block for the subnet.

    Returns:
        None

    """
    client.subnets.begin_create_or_update(
        resource_group_name=resource_group_name,
        virtual_network_name=vnet_name,
        subnet_name=subnet_name,
        subnet_parameters={"address_prefix": cidr},
    ).result()


def create_virtual_networks(
    credential: DefaultAzureCredential,
    subscription_id: str,
    locations: list[str],
    resource_group_name: str,
    rich_progress: Progress | None = None,
    rich_task: TaskID | None = None,
) -> list[str]:
    """Create virtual networks in the specified locations.

    Args:
        credential (DefaultAzureCredential): Azure credentials for authentication.
        subscription_id (str): Azure subscription ID.
        locations (list[str]): List of Azure locations to create virtual networks in.
        resource_group_name (str): Name of the resource group to create the virtual networks in.
        rich_progress (Progress | None, optional): Optional Rich Progress instance for progress tracking. Defaults to None.
        rich_task (TaskID | None, optional): Optional task ID for progress tracking. Defaults to None.

    Returns:
        list[str]: List of virtual network names created.

    """  # noqa: E501
    network_client = NetworkManagementClient(credential, subscription_id)
    networks = []
    for location in locations:
        if rich_progress and rich_task is not None:
            rich_progress.update(
                rich_task,
                description=f"Creating Virtual Network: {resource_group_name}-{location}",
                advance=1,
            )
        virual_network_cidr, pod_network_cidr, service_network_cidr = generate_cidrs()
        vnet = create_vnet(
            network_client,
            resource_group_name,
            location,
            virual_network_cidr,
            pod_network_cidr,
            service_network_cidr,
        )
        networks.append(vnet.name)
        subnets = [str(network) for network in ip_network(virual_network_cidr).subnets(new_prefix=24)]
        create_subnet(network_client, resource_group_name, vnet.name, "kube_nodes", subnets[0])
    if len(networks) > 1:
        peer_vnets(network_client, subscription_id, resource_group_name, networks)
    return networks



"""Initialization module for the Kina CLI."""

from typing import Annotated

import typer
from azure.identity import DefaultAzureCredential
from rich.console import Console
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rich.table import Table

from kina.azure.helpers import get_default_subscription_id

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)
kubectl = typer.Typer(no_args_is_help=True, help="Manage local Kubectl Configuration")
app.add_typer(kubectl, name="kubectl")

credential = DefaultAzureCredential()
subscription_id = get_default_subscription_id()


@app.command(name="create")
def cluster_create(
    locations: Annotated[
        str,
        typer.Option(
            "--locations",
            "--location",
            "-l",
            help="Azure Locations to create Kubernetes Clusters in, comma seperated",
        ),
    ] = "uksouth",
    no_cni: Annotated[
        bool,
        typer.Option(
            "--no-cni",
            help="Disable CNI for the AKS clusters. This is useful for testing purposes.",
        ),
    ] = False,
) -> None:
    """Create a new Kina AKS Cluster.

    Args:
        locations (str): Azure Locations to create Kubernetes Clusters in, comma separated.
        no_cni (bool): Disable CNI for the AKS clusters. This is useful for testing purposes.

    """
    from kina.azure.kubernetes import configure_network_iam, create_aks_clusters
    from kina.azure.resource_groups import create_resource_group
    from kina.azure.virtual_networks import create_virtual_networks
    from kina.kubectl import add_cluster_to_kubeconfig, set_current_context

    locations = locations.split(",")
    cluster_count = len(locations)

    with Progress(
        SpinnerColumn(),
        BarColumn(),
        TimeElapsedColumn(),
        TextColumn("[progress.description]{task.description}"),
    ) as progress:
        rg_task = progress.add_task("Creating Resource Group...", total=1)
        vnet_task = progress.add_task("Creating Virtual Networks...", total=cluster_count + 1)
        aks_task = progress.add_task("Creating AKS Clusters...", total=cluster_count + 1)
        kubectl_task = progress.add_task("Adding AKS Cluster to Kubeconfig...", total=cluster_count + 1)

        rg_name = create_resource_group(
            credential=credential,
            subscription_id=subscription_id,
            locations=locations,
        )
        progress.update(rg_task, description=f"Created Resource Group: {rg_name}", advance=1)
        vnets = create_virtual_networks(
            credential=credential,
            subscription_id=subscription_id,
            locations=locations,
            resource_group_name=rg_name,
            rich_progress=progress,
            rich_task=vnet_task,
        )
        progress.update(vnet_task, description=f"Created Virtual Networks: {', '.join(vnets)}", advance=1)
        clusters = create_aks_clusters(
            credential=credential,
            subscription_id=subscription_id,
            virtual_network_names=vnets,
            resource_group_name=rg_name,
            no_cni=no_cni,
            rich_progress=progress,
            rich_task=aks_task,
        )
        for cluster in clusters:
            configure_network_iam(
                credential=credential,
                subscription_id=subscription_id,
                resource_group=rg_name,
                cluster_name=cluster,
            )
        progress.update(aks_task, description=f"Created AKS Clusters: {', '.join(clusters)}", advance=1)
        for cluster in clusters:
            progress.update(kubectl_task, description=f"Adding AKS Cluster to Kubeconfig: {cluster}", advance=1)
            add_cluster_to_kubeconfig(
                credential=credential,
                subscription_id=subscription_id,
                resource_group_name=rg_name,
                cluster_name=cluster,
            )
        set_current_context(clusters[0])
        progress.update(
            kubectl_task,
            description=f"Added AKS Clusters to Kubeconfig: {', '.join(clusters)}, set current context to {clusters[0]}",  # noqa: E501
            advance=1,
        )


@app.command(name="delete")
def cluster_delete(name: Annotated[str, typer.Argument()]) -> None:
    """Delete an existing Kina AKS Cluster.

    Args:
        name (str): The name of the Kina AKS Cluster to delete.

    """
    from kina.azure.resource_groups import delete_resource_group

    deleted = delete_resource_group(credential, subscription_id, name)
    if deleted:
        typer.echo(f"Deleted Resource Group: {name}")
    else:
        typer.echo(f"Resource Group '{name}' not found or not managed by Kina.")


@app.command(name="list")
def cluster_list(
    output: Annotated[
        str,
        typer.Option("--output", "-o", help="Output format. Allowed values: json, table, names"),
    ] = "table",
) -> None:
    """List all Kina AKS Clusters.

    Args:
        output (str): Output format. Allowed values: json, table, names.

    """
    from kina.azure.resource_groups import list_resource_groups

    resource_groups = list_resource_groups(credential, subscription_id)
    console = Console()

    if output == "table":
        t = Table()
        t.add_column("Name")
        t.add_column("Location(s)")
        t.add_column("Created By")

        for resource_group_name, locations, username in resource_groups:
            t.add_row(resource_group_name, locations, username)
        console.print(t)
    elif output == "json":
        import json

        resource_groups_json = [
            {
                "name": resource_group_name,
                "locations": locations,
                "created_by": username,
            }
            for resource_group_name, locations, username in resource_groups
        ]
        console.print(json.dumps(resource_groups_json, indent=2))
    elif output == "names":
        resource_group_names = [resource_group_name for resource_group_name, _, _ in resource_groups]
        console.print("\n".join(resource_group_names))


@kubectl.command()
def cleanup() -> None:
    """Clean up the kubectl configuration by removing all Kina instances."""
    from kina.kubectl import cleanup_kubeconfig

    cleanup_kubeconfig()
    typer.echo("Cleaned up kubectl configuration")


@kubectl.command()
def remove(name: str) -> None:
    """Remove a Kina instance from the kubectl configuration.

    Args:
        name (str): The name of the Kina instance to remove.

    """
    from kina.kubectl import remove_from_kubeconfig

    remove_from_kubeconfig(name)
    typer.echo(f"Removed Kina Instance from Kubeconfig: {name}")


if __name__ == "__main__":
    app()
